program Life(input,output);          (* version: 23 Aug 2005 *)

(*  Trace the life history of a colony of cells, generation by generation.
  The birth, survival, or death of each cell is determined by the number
  of adjacent cells.  The math behind "the game of life" is by John Conway,
  see Martin Gardner's column in Scientific American 223 (October 1970): 120-123.  *)
(*  This program is in the public domain; beginning from "life" for Apple II
  in "Apple Pascal Games", by Douglas Hergert and Joseph T. Kalash, 1981; 
  heavily modified and extended by W. Kempton.  Most code is ISO standard Pascal. *)
(* Currently compiled by GPC  Pascal on Mac OSX.  ncurses used for function
  keys and screen.  This version serves as a prototype for writing programs which
  use ncurses function keys, with  key definitions through table initialization.
  gpc compile with:   gp life
 *)

import ncurses;                       { ncurses pascal interface }
       GPC only (SleepMicroSeconds);  { import SleepMicroSeconds from GPC }
       ncursesPlus;                   { HasKeySet, KeyLabels }

const
  { program options (controlled via constants) }
  ForceASCII = false;    (* if true, do not use ncurses/vt102 alt char set *)
  InterGenDelay = 12;    (* how many 1/10 sec wait to see each generation; 12 recommended *)
  RepeatsToEnd = 15;     (* how many repeated (population and edges) to stop? (0 disables) *)
  DeadLeaveTrace = false; (* if true, small mark to show prior living cells *)

  { size constants }
  maxheight = 60   (* max lines on screen, plus one *);
  maxwidth  =110   (* max columns on screen, plus one *);
  minbound  = -1   (* minimum line and column, minus one; ANSI is 0, ncurses is -1 *);
  kTabMax  = 22;   (* max entries in key table *)
  kStringMax=10;
  kStrTabMax=200; (* table of key codes and names *)

type
  CellState = (alive, dead);
  GameStatus = (progressing, died, stable, repeating);
  DetectIndexType = 0..5;

  cell = record
    LooksLikeItIs : CellState   (* is this position alive or dead? *);
    nearby        : integer (* count of living neighbors *);
  end;

  edges = record
    (* the coordinates of a square which would contain all living cells *)
    (* this is used to limit search area, thus speeding execution *)
    left, right : minbound .. maxwidth;
    top, bottom : minbound .. maxheight;
  end;

  RepeatRec = record  { for detecting repeating loops }
      pop: integer; edg: edges; 
  end;

  commands = (nocommand, help, right, left, down, up, select, startrun,
              interrupt, exit, badcommand, all, PartialCommand,OneCharCommand);
  SetOfCommands = set of commands;
  keystrings = packed array[1..kStringMax] of char;
  KeyStrTabIndx = 0..kStrTabMax;

  keyentry = record
    (* entry in table of funtion keys on this machine *)
    kfun: commands;       (* command that this kcode string denotes *)
    kcode,                (* chars generated by key *)
    kname: KeyStrTabIndx; (* label printed on key *)
  end;

  ttabentry = record case boolean of
      true: (i:integer);
      false:(c:char);
   end;

  kRecordString = record
    len: integer;
    c  : keystrings;
  end;

var
  board : array[minbound..maxheight] of array [minbound..maxwidth] of cell;
  width, height: integer;  { screen coordinates, could be minbound..maxwidth/maxheight }
  population, births, deaths : integer;
  ch : char;
  edge : edges;
  vt102: boolean; (* a vt102, vt220, etc, or  Rainbow 100 *)
  ICH:  chtype;  (* global current input char *)
  AliveChar,     (* use this character--if cell now alive *)
  DeadChar ,     (* use this character--if cell was alive here previously *)
  NeverAliveChar (* this character is left on screen if cell was never alive  *)
    : chtype; { ncurses type }
  KeyBeginCodes, KeyEndCodes: set of char;
  FreeKeyEntry: integer;   (* after key entry actually entered ( < kTabMax) *)
  FreeKeyStr: KeyStrTabIndx; (* after key string *)
  ktab: array [0..kTabMax] of keyentry;
  PendingCommand: commands;
  kStrTab: packed array [KeyStrTabIndx] of ttabentry; (* strings=codes+names *)
  GameState: GameStatus;
  RepeatDetect: array[0..4] of RepeatRec;
  DetectIndex: DetectIndexType;
  RepeatCount: integer;
  retcode: Curses_int; { usually a dummy for return value (from ncurses functions) }

procedure enter(c: commands; kc, kn: keystrings);
(* add one more entry to the key table *)
(*  this is an OS-generic function key entry, not dependent on ncurses *)
var i, j, length : integer;
begin
  if FreeKeyEntry> kTabMax then write(output,'key overflow');
  if kc[1]='\' then kc[1]:= chr(27); (* make escape typeable *)
  KeyBeginCodes := KeyBeginCodes + [ kc[1] ];
  with ktab[FreeKeyEntry] do
    begin
      kfun:= c;
      (* first enter key code, then key name *)
     for j := 0 to 1 do begin  if j=1 then kc :=kn;
      length := kStringMax;
      while kc[length] = ' ' do length:=length-1;
      if j=0 then kcode := FreeKeyStr else kname := FreeKeyStr;
      kStrTab[FreeKeyStr].i := length;
      for i := 1 to length do  kStrTab[FreeKeyStr+i].c := kc[i];
      FreeKeyStr := FreeKeyStr + length + 1;
      if j=0 then KeyEndCodes := KeyEndCodes + [ kc[length] ];
     end;

    end;
  FreeKeyEntry := FreeKeyEntry + 1;
end (* enter *);

procedure enterf(c: commands; funckey: chtype);
{ Enter an ncurses function key rather than strings. }
{ For generic key entry, use enter or CtrlEnter, not enterf. }
var kc, kn: keystrings;

begin { enterf }
  if (funckey < KEY_CODE_YES) or (funckey > KEY_CODE_YES + 255) then halt;
  kc := '          ';    kn := '          ';
  kc[1] := chr(255); kc[2] := chr(funckey-255);  
  kn := KeyLabels[funckey];  { from ncursesPlus }
  enter(c, kc, kn);
end { enterf };


procedure InitKeys;
(* stick keycodes into global array.  note: enter uses '\' to denote escape *)
(* in case of duplicate keys for one function, the name of the last-entered
 will be used in instructions and help menus. *)
(* This is calls both the generic key-entry routine, and the ncurses-dependent one. *)
var kc,kn: keystrings;

  procedure CntlEnter(comm:commands; c:char);
  begin kc[1] := chr(ord(c)-64); kn[2] := c; enter(comm,kc,kn);
  end;

begin
  KeyBeginCodes :=[];  KeyEndCodes := [];
  (* sentenals put in:     kStrTab[0].i   and     ktab[0].kfun *)
  FreeKeyStr := 1;   FreeKeyEntry := 1;  ktab[0].kcode:=0;

  (* first, enter generic (WordStar-type) control characters *)
  kc := '          ';  kn := '^         ';
  CntlEnter(up,      'E');  CntlEnter(left,    'S');  CntlEnter(right,   'D');
  CntlEnter(down,    'X');  CntlEnter(select,  'V');  CntlEnter(exit,    'K');
  CntlEnter(StartRun,'R');  CntlEnter(help,    'Y'); CntlEnter(interrupt,'U');
  { Enter movement keys, assuming ncurses s available }
  enterf(up, KEY_UP);
  enterf(down, KEY_DOWN);
  enterf(right, KEY_RIGHT);
  enterf(left,  KEY_LEFT);
  { enter several synonyms for functional keys, if they are in terminfo; last entered will display }
  if KEY_F1 in HasKeySet   then enterf(help, KEY_F1);
  if KEY_HELP in HasKeySet then enterf(help, KEY_HELP);
  if KEY_BACKSPACE in HasKeySet then enterf(select,KEY_BACKSPACE); 
  if KEY_F2 in HasKeySet then enterf(select,KEY_F2);
  if KEY_SELECT in HasKeySet then enterf(select,KEY_SELECT);{ select is used to add or delete }
  if KEY_BEG in HasKeySet then enterf(startrun,KEY_BEG);
  if KEY_F3 in HasKeySet then enterf(startrun,KEY_F3);
  if KEY_REDO in HasKeySet then enterf(startrun,KEY_REDO);
  if KEY_CANCEL in HasKeySet then enterf(interrupt,KEY_CANCEL);
  if KEY_F4 in HasKeySet then enterf(interrupt,KEY_F4);
  if KEY_BREAK in HasKeySet then enterf(interrupt,KEY_BREAK);
  if KEY_END in HasKeySet then enterf(exit,KEY_END);
  if KEY_F5 in HasKeySet then enterf(exit,KEY_F5);
  if KEY_EXIT in HasKeySet then enterf(exit,KEY_EXIT);

 if vt102 then
 begin
  (* if ncurses not working, use following for vt220, DEC Rainbow *)
  enter(up,       '\[A       ','Up arrow  ');
  enter(down,     '\[B       ','Down arrow');
  enter(right,    '\[C       ','-->       ');
  enter(left,     '\[D       ','<--       ');
  enter(select,   '\[4~      ','Select    ');
  enter(StartRun, '\[29~     ','Do        ');
  enter(interrupt,'\[17~     ','Interrupt ');
  enter(help     ,'\[28~     ','Help      ');
  enter(exit,     '\[21~     ','Exit      ');
 end;
end;

(* ************* all nonportable code is below this line **************** *)

{$ignore-function-results}  { this avoids need for retcode:= on most ncurses library calls }

procedure getchq;
(* input into global character, without echo to console *)
begin
  ICH := getch;
  { read(kbd,ICH); } (* TURBO: kbd means no echo *)
end;

function KeyPressed: boolean; 
{ Return true if a key has been pressed.  Do not get key. }
{ clumsy: via ncurses, we must make two calls to nodelay, and }
{   then push the character back in the queue }
var chi, retcode: Curses_int; { the ncurses type returned by getch }
begin
  retcode := nodelay(stdscr, true);
  chi := getch;
  retcode := nodelay(stdscr, false);
  if chi = Curses_ERR
    then KeyPressed := false
    else begin KeyPressed := true;  retcode := ungetch(chi);  end;
end { KeyPressed };

procedure Wait(n: integer);
(* wait n tenths of a second *)
begin
  SleepMicroSeconds(n * 10000); { GPC call} { 10,000 microseconds = 1/10 second }
  { could also use curses "napms": nap milliseconds }
end;

procedure BeepOrFlash;
(* screen flash  if possible, ASCII bell ring for the rest *)
begin
  flash; { ncurses: flash if possible, otherwise beep }
end;

procedure ClearScreen;
(* clear the entire screen  *)
begin
   clear;  { ncurses }
   { write(output,chr(27),'[2J'); } { ANSI }
end;

procedure header;  { NOT USED in ncurses version }
(* code is:  esc, poundsign, 3 for top; esc, poundsign, 4 for bottom *)
var i: 3..4;
begin
  if not vt102
    then writeln(' ':((width-4)div 2),'LIFE')
    else
     for i:= 3 to 4 do
      writeln(output,chr(27),chr(35), i:1,
            ' ':(((width div 2)-4) div 2), (* double width, center *)
             'LIFE');
end;


procedure AddInt(cc: integer);
{ puts up to a three-digit integer in the ncurses output buffer }
begin
  if cc>9 then begin
    if cc>99 then begin
        addch(ord('0') + cc div 100);
         cc :=  cc mod 100;
      end;
     addch(ord('0') + (cc div 10) );
     cc := cc mod 10;
   end;
   addch(ord('0') + (cc ) );
end { AddInt };

procedure gotoxy(x,y: Curses_int {integer} );
{ go to position x,y --  column x, line y }
{ ncurses move (and ANSI code) is ordered line, col; Turbo built-in gotoxy is col, line } 
begin
   move(y,x);
   { write(output,chr(27), '[', y:1,';', x:1, 'f');  } { ANSI terminal }
end (* gotoxy *);


procedure MachineInitialize;
(* any nonstandard initializations  *)
type
  s30= string(30);
var
  linecount, colcount: Curses_int; { type cint from ncurses.pas  }
  i: integer;

begin
  { initialize ncurses, prep function keys }
  initscr;
   noecho;
  { cbreak; }    { No line buffering, pass on as typed including erase, but process ^C and ^Z }
   raw;   { raw additionally does not interpret interrupt, quit, suspend, and flow control  characters }
   keypad(stdscr, true);  { enable ncurses' recognition of F keys }
  { use ncurses acs (alternate character set, unless ForceASCII }
  NeverAliveChar := ord(' ');
  if ForceASCII 
     then begin AliveChar := ord('*');    DeadChar := ord('-');   end
     else 
       begin AliveChar := ACS_DIAMOND;  { others options:  ACS_CKBOARD }
             DeadChar := ACS_BULLET; 
       end;
  If not DeadLeaveTrace then DeadChar := NeverAliveChar;

  ICH := ord(' ');  { be sure this is defined, before some routine tries to use it }
  pendingcommand := nocommand; { no command waiting }

  vt102 := false;
   {  getbegyx(stdscr, linecount, colcount); } { not used, assume global constant "min" is right }
   { first, get screen limits }
   getmaxyx(stdscr, linecount, colcount);
   { limit the max line, col to the internal array size.  }
   if linecount >= maxheight then linecount := maxheight;
   if colcount >= maxwidth   then colcount  := maxwidth;
   height := linecount; { assumes min=0 }
   width  := colcount ; { assumes min=0 }
   { move(2,0); addstr('height, width = ' +  IntToStr(height) + ', ' + IntToStr(width)); refresh;}
   { getchq; } { pause for user to see }
end { MachineInitialize };

procedure MachineCloseup;
(* any nonstandard cleanup or closeout after the above MachineInitialize *)
begin
  endwin;  { ncurses }
end;


{$no-ignore-function-results}

(* *************  all nonportable code is above this line  ************ *)


(*  ************** user input routines ************** *)

function keyfunction(kc: kRecordString): commands;
(* look up the keystroke sequence, and return the command *)
var kindex: 0..kTabMax; ksindex: KeyStrTabIndx;
    exit: boolean; length, i: integer;
begin
  if not(kc.c[1] in KeyBeginCodes)
    then KeyFunction := OneCharCommand
    else
      begin
        length := kc.len;
        kStrTab[ktab[0].kcode].i:= length; (* sentenal *)
        ktab[0].kfun := badcommand;  (* if no match *)
        kindex := FreeKeyEntry;
        repeat
          repeat (* find one the right length *)
             kindex := kindex - 1; ksindex := ktab[kindex].kcode;
          until length <= kStrTab[ksindex].i;
          exit := true;
          if kindex > 0 then
            for i := 1 to length do
              if kStrTab[ksindex+i].c <> kc.c[i] then exit := false;
        until exit;
        if length = kStrTab[ksindex].i
          then keyfunction := ktab[kindex].kfun
          else keyfunction := PartialCommand;
      end;
end (* keyfunction *);

procedure getkey(var kcode: kRecordString);
(* get a single character, or a command sequence, from the keyboard *)

   procedure nextkey;  (* wait for a single character *)
    begin with kcode do
      begin 
         getchq;   len:=len+1;  c[len]:=chr(ICH);
      end;
    end;

begin (* getkey *)
  kcode.len := 0;
  repeat 
    getchq;   
    kcode.len := kcode.len +1;
    with kcode do
      if ICH < KEY_CODE_YES 
         then begin kcode.c[len] := chr(ICH);  end
         else begin kcode.c[len] := chr(255); len:= len+1; kcode.c[len] := chr(ICH-255); end;
  until keyfunction(kcode) <> PartialCommand;
end (* getkey *);


procedure KeyLabel(kf: commands);
(* Write the name of the key.  Seach the table from top down, writing only
  the first name found. *)
var k, i: integer;  retcode: Curses_int;
begin
  k:=FreeKeyEntry-1; ktab[0].kfun := kf;
  while (ktab[k].kfun <> kf) do k:=k-1;
  if (ktab[k].kfun <> kf)
   then retcode:= addstr('    ') (* not found: nothing is self-explanatory to write *)
   else
    with ktab[k] do begin (* write without trailing blanks *)
       retcode := attrset(A_REVERSE);
       { write(chr(27),'[7m'); } (* ANSI reverse *)
       for i:= 1 to kStrTab[kname].i (* length *) do
          retcode := addch(ord(kStrTab[kname+i].c));
       retcode := attroff(A_REVERSE);
       { write(chr(27),'[0;1m');}  (* ANSI no reverse, bold *)
     end;
   retcode := refresh;
end (* KeyLabel *);

function nextcommand(allowed: SetOfCommands): commands;
(* return the pending command, or a new one from the keyboard *)
var 
   kc: kRecordString; patience, i : integer;  
   c: commands;  MessageRequired: boolean;
   retcode: Curses_int; { ncurses return value }

  function ok(c:commands): boolean;
  begin
     ok := (c in allowed) or
     ( (allowed = [all]) and not(c in [badcommand,nocommand]));
  end (* ok *);

begin (* nextcommand *)
  patience := 3;
  MessageRequired := false;
  if  ok(pendingcommand)
    then begin nextcommand:=pendingcommand; pendingcommand:=nocommand end
    else
      begin
      repeat
        getkey(kc);  c:=keyfunction(kc); nextcommand := c;
        if not ok(c) then
          begin
            BeepOrFlash; patience := patience - 1;
            if patience <= 0 then
              begin
                   gotoxy( (minbound+1), (height-2));
                   MessageRequired := true;
                   retcode := clrtoeol;
                   retcode := addstr('Allowed keys are: ');
                   for c:= nocommand to badcommand do
                     if ok(c) then 
                       begin KeyLabel(c); retcode:=addch(ord(' ')); end;
                   retcode := refresh;
                   c := badcommand; (* insure correct loop exit *)
              end;
         end;
      until ok(c);
      if c = OneCharCommand then ICH := ord(kc.c[1]);
      if MessageRequired then begin 
          gotoxy( (minbound+1), (height-2));
          retcode := clrtoeol;
        end;
      end;
end (* nextcommand *);


(* ************** utilities ************** *)

procedure ResetEdges;
(*  Initialize the edges of the pattern.  Initialized to guarantee these
  values will be replaced by anything.  *)
begin
  with edge do
    begin  top := height-1;  right := minbound+1;
           left:= width -1;  bottom:= minbound+1;
    end;
end (* ResetEdges *);

function max (a,b: integer): integer;
(* return larger argument *)
begin
  if  a >= b then  max:=a  else  max := b;
end (* max *);


function min (a,b: integer): integer;
(* return smaller argument *)
begin
  if a <= b then   min:=a  else  min:=b;
end (* min *);

procedure limits(x, y: integer);
(* Determine if and how the coordinates passed as argument change the
  bounds of the pattern (the box containing all living cells), checking
  that it does not go off one of the sides of the board.  *)
begin
  with edge do
  begin
    left  := min(left,x);
    right := max(right,x);
    top   := min(top,y);
    bottom:= max(bottom,y);
  end;
end (* limits *);


(* ************** initialization ************** *)


procedure instructions;
{ write instructions to the user }
type str54= packed array[1..54] of char;
var retcode: Curses_int; { ncurses return code }
    lin: 0..255;

  procedure nextlin;
   begin
     lin := lin + 1;
     gotoxy(0,lin);
   end;

  procedure interp(s: str54);
  var n: 0 .. 54;
  begin
    for n := 1 to 54 do
     if s[n] = ' '
        then retcode := addch(NeverAliveChar)
        else retcode := addch(AliveChar);
     nextlin;
  end { interp };

begin
      lin := 3;
      nextlin;
      retcode := addstr('Life simulates the growth of a colony of animals on a ');
      AddInt(height-minbound-1); retcode := addstr(' row by '); 
      AddInt(width-minbound-1); retcode := addstr(' column ' ); 
      nextlin;
      retcode := addstr('world.  Whether a cell is born, lives, or ');
      retcode := addstr('dies depends on the number ');
      nextlin;
      retcode := addstr('of living animals adjacent to it.  If ');
      retcode := addstr('a cell is empty and has exactly three ');
      nextlin;
      retcode := addstr('neighbors, it will be born in the ');
      retcode := addstr('next generation.  If it is alive, then ');
      nextlin;
      retcode := addstr('if it has either two or three neighbors it ');
      retcode := addstr('will continue to live, otherwise ');
      nextlin;
      retcode := addstr('it will die of loneliness or overcrowding.');
      nextlin;
      nextlin;
      retcode := addstr('To enter the initial pattern, move the ');
      retcode := addstr('cursor keys to each position, then ');
      nextlin;
      retcode := addstr('use the '); KeyLabel(select);
      retcode := addstr(' key to create a cell there.  When ready, use the ');
              KeyLabel(startrun);
      retcode := addstr(' key to ');
      nextlin;
      retcode := addstr('begin.  A running pattern may be stopped by ');
      retcode := addstr('typing ');
              KeyLabel(interrupt);
      nextlin;
      nextlin;
      retcode := addstr(' A few interesting initial patterns are ');
      retcode := addstr('shown below: ');
      nextlin;
      nextlin;
      interp('                                                ');
      interp('                                                ');
      interp('                                             ');
      nextlin;
    (* ten cells in a row is a good long-cycle repeater *)

      retcode := addstr(' Type any key to start entering your pattern: ');
      retcode := refresh;
      if nextcommand([all]) = help then (* do nothing *) ;
end (* instructions *);


procedure InitializeBoard;
(*  Reset the board to empty (all dead and no neighbors).  *)
var
  down, across : integer;
begin
  for down := minbound to height do
    for across := minbound to width do
    with  board[down,across] do
      begin
        LooksLikeItIs := dead;  nearby := 0;
      end;
  ResetEdges;
  DetectIndex := 0;
  RepeatCount := 0;
end (* InitializeBoard *);


function getone(var y, x: integer): boolean;
(* get one position, using function keys *)
 var  done: boolean;
begin
   getone := true;
   done := false;
       repeat
         gotoxy(x,y); (* show current cursor position *)
         case nextcommand([right,left,up,down,select,startrun]) of
           right   : if (x+1)>=width    then BeepOrFlash else x := x+1;
           left    : if (x-1)<=minbound then BeepOrFlash else x := x-1;
           up      : if (y-1)<=minbound then BeepOrFlash else y := y-1;
           down    : if (y+1)>=height   then BeepOrFlash else y := y+1;
           select  : done := true;
           startrun: begin done := true; getone := false;  end;
         end (* case *);
       until done;
end (* getone *);

procedure getpositions;
(*  Get the initial live cells from the user.  *)
var
  down, across : integer;
  finished : boolean;
  retcode: Curses_int; { error code return from ncurses }

begin
  finished := false;    population := 0;
  ClearScreen;
  down := (height div 2); across := (width div 2); (* initial values *)

    while getone(down, across) do

      with board[down,across] do
      begin
        limits(across, down);
        gotoxy(across,down);
        if LooksLikeItIs = alive
         then (* already something there--erase *)
           begin
            retcode := addch(NeverAliveChar);
            LooksLikeItIs := dead;
            population := population -1;
           end
         else
           begin
            retcode := addch(AliveChar);
            LooksLikeItIs := alive;
            population := population + 1;
           end;
      end;
end (* GetPositions *);

procedure initgame;
(* offer instructions, then get initial configuration *)
var retcode: Curses_int;
begin
   ClearScreen;
  gotoxy((width div 2)-4, 1); retcode := addstr(' L I F E '); 
  gotoxy(0,5); retcode := addstr('Type ');
  KeyLabel(help);
  retcode := addstr(' for instructions, any other key to begin.');
  retcode := refresh; 
  wait(10);  { TEMPORARY DEBUG }
  InitializeBoard;
  if nextcommand([all]) = help then begin gotoxy(0,5); instructions; end;
  clearscreen;
  getpositions;
end (* initgame *);


(* ************** play the game of life ************** *)



procedure ClearNearby;
(*  Zero the neighbors of all cells, to prepare calculations for next
  generation.  *)
var
  down, across : integer;
begin
  for down := edge.top - 1 to edge.bottom +1 do
    for across := edge.left - 1  to  edge.right + 1  do
      board[down,across].nearby := 0;
end (* ClearNearby *);


procedure CountNeighbors;
(*  Compute the number of adjacent cells, and thus which cells will survive
  through the next generation.  To simplify the code, the middle cell of the
  3 by 3 area being examined is included in the count, even though it is
  not really a neighbor of itself.  This off-by-one discrepancy is adjusted
  for in the board update.  *)
var
  down, across : integer;
  LocalDown, LocalAcross : integer;
begin
  ClearNearby  (* zero count from previous generation *);
  for down :=  edge.top-1  to  edge.bottom+1  do
    for across :=  edge.left-1  to  edge.right+1  do
      if board[down,across].LooksLikeItIs = alive then
        for LocalDown := down-1 to down+1 do
          for LocalAcross := across-1 to across+1 do
            with board[LocalDown,LocalAcross] do
              nearby := nearby + 1;
end (* CountNeighbors *);


procedure update;
(*  Shows births and deaths on the screen *)
var
  down, across : integer;
  localedge : edges;
  retcode: Curses_int; { return code from ncurses }
begin
  births := 0;    deaths := 0;
  localedge := edge;
  ResetEdges;

  for down := max(minbound+1, localedge.top-1) to
                             min(height-1, localedge.bottom+1) do
    for across := max(minbound+1, localedge.left-1) to
                               min(width-1, localedge.right+1) do
      with board[down,across] do

        case LooksLikeItIs of
         
         dead:
               if nearby = 3 then
               begin
                 LooksLikeItIs := alive;
                 gotoxy(across,down);
                 retcode := addch(AliveChar);
                 limits(across,down);
                 births := births + 1;
               end;

         alive:
               if (nearby = 3) or (nearby = 4)
                 then limits(across,down)
                 else
                   begin
                     LooksLikeItIs := dead;
                     gotoxy(across,down);
                     retcode := addch(DeadChar);
                     deaths := deaths + 1;
                   end;
         end (* case *);
    gotoxy(0,0); (* wait at upper left during computations *)
    population := population + births - deaths;
end (* update *);

function gameover: boolean;
begin
  gameover := (population = 0) or ((births = 0) and (deaths = 0));
end;

procedure FindGameState(var s: GameStatus);
{ Figure out if the colony has died, is repeating or is progressing }

  function equalpat(a,b:RepeatRec): boolean;
   begin
     equalpat :=
       (a.pop = b.pop) and
       (a.edg.top    = b.edg.top) and
       (a.edg.bottom = b.edg.bottom) and
       (a.edg.left   = b.edg.left ) and
       (a.edg.right  = b.edg.right);
   end;

begin
  DetectIndex := DetectIndex mod 5; { runs 0..4 and repeats } 
  with RepeatDetect[DetectIndex] do 
    begin pop:=population; edg:= edge; end;
  DetectIndex := DetectIndex + 1;
  if DetectIndex = 5 then
       if equalpat(RepeatDetect[4],RepeatDetect[2]) and
          equalpat(RepeatDetect[4],RepeatDetect[0]) 
        then RepeatCount := RepeatCount + 1
        else RepeatCount := 0;
  s := progressing;
  if population = 0 then s := died
  else if (births =0) and (deaths=0) then s := stable
  else if (RepeatsToEnd<>0) and (RepeatCount>RepeatsToEnd) then s := Repeating; 
end { FindGameState };

begin  (*  LIFE  *)
  MachineInitialize;
  InitKeys;
 repeat  (* multiple games restart here *)
   InitGame;
   repeat
     pendingcommand := badcommand;
     repeat
       CountNeighbors;
       update;
       retcode := refresh; { refresh screen }
       FindGameState(GameState);
       wait(InterGenDelay); { tenths of a second wait after each generation }
     until (GameState<>progressing) or KeyPressed;
     if keypressed then pendingcommand := nextcommand([all]);
   until (GameState<>progressing) or (pendingcommand in [exit, interrupt]);

  (* following position prevents CLI prompt from scrolling the display *)
  gotoxy( (minbound+1), (height-2));
  if (pendingcommand in [exit, interrupt])
   then retcode := addstr('Ended by user')
   else
    begin
    case GameState of
      died:     retcode := addstr('This colony has died');
      stable:   retcode := addstr('The pattern is stable');
      repeating: retcode := addstr('The colony keeps repeating');
      progressing:  { presumably 'ended by user' } ;
    end { case };
    end;
   retcode := refresh;
   Wait(30);

  if pendingcommand <> exit then
   begin pendingcommand := nocommand;
      retcode := addstr( '; ');  KeyLabel(exit);
      retcode := addstr(' to exit or ');
      KeyLabel(StartRun); retcode := addstr(' to restart ');
      retcode := refresh;
   end;
 until nextcommand([exit,StartRun]) = exit;
 MachineCloseup;
end (* life *).
